import json
from datetime import datetime
from typing import List, Optional, Iterable

import requests.exceptions
from loguru import logger
from requests_html import HTMLSession, HTML

from .crawler import Crawler
from ..config import g1_api

SESSION = HTMLSession()


class G1Comments(Crawler):
    _COMMENTS_API: str
    _API_CONFIG: dict

    def __init__(self) -> None:
        self._API_CONFIG = g1_api.comments_config

        self._COMMENTS_API = self._API_CONFIG["api_url"]["comments_engine"]
        self._COUNT_API = self._API_CONFIG["api_url"]["count_engine"]
        self._API_PARAMS = self._API_CONFIG["params"]
        self._ERRORS = (
            requests.exceptions.ReadTimeout, requests.exceptions.InvalidSchema, requests.exceptions.MissingSchema
        )

    def _news_have_comments(self, target_url: str) -> bool:
        response = SESSION.get(url=f"{self._COUNT_API}{target_url}")
        if response.status_code != 200:
            return False

        response_data = response.json()

        count_data = int(response_data["count"]) if response_data["count"] is not None else None
        if count_data is None:
            return False
        if count_data == 0:
            return False

        return True

    def _get_comment_data(self, news_url: str) -> Optional[List[dict]]:
        if self._news_have_comments(target_url=news_url) is False:
            return None

        parameters = dict(self._API_PARAMS)
        parameters["variables"] = str(parameters["variables"].replace("@", news_url))

        response = SESSION.get(url=self._COMMENTS_API, params=parameters)
        if response.status_code != 200:
            logger.error(f"Error while getting comments from {news_url} @ Status Code: {response.status_code}")
            return None

        try:
            news_data = response.json()['data']['story']['comments']['edges']
        except (json.decoder.JSONDecodeError, TypeError):
            news_data = None

        if news_data is not None:
            return news_data

        return None

    def parse_comments(self, news_list: List[dict]) -> Iterable[dict]:
        for news_data in news_list:
            url = news_data["url"]

            raw_data = self._get_comment_data(news_url=url)
            if raw_data is None:
                continue

            for comment_node in raw_data:
                node_data = comment_node["node"]
                if len(node_data["body"]) == 0:
                    continue

                data = {
                    "author": node_data["author"]["username"],
                    "date": datetime.strptime(node_data["createdAt"], "%Y-%m-%dT%H:%M:%S.%fZ"),
                    "upvote": int(node_data["actionCounts"]["reaction"]["total"]),
                    "news_data": {
                        "title": news_data["title"],
                        "region": news_data["region"],
                        "url": url,
                    },
                    "comment": HTML(html=node_data["body"]).full_text,
                    "g1_id": node_data["id"],
                    "platform": "G1",
                }

                yield data



import csv
import json
from abc import ABC, abstractmethod
from datetime import datetime


class Crawler(ABC):
    @abstractmethod
    def parse_comments(self, news_urls: list):
        pass

    @staticmethod
    def export_data(parsed_data: list, export_type: str = 'csv'):
        export_time = datetime.today().strftime('%Y_%m_%d_%HH_%MM')
        if len(parsed_data) == 0:
            raise ValueError(f'parsed_data ({export_type}) cannot be empty.')
        if ('csv' not in export_type.lower()) and ('json' not in export_type.lower()):
            raise ValueError(f'{export_type} is not a valid export format.')
        with open(f'ParsedNewsData_{export_time}.{export_type}', 'w', encoding='utf-8') as export_file:
            if export_type.lower() == 'csv':
                header = ['title', 'abstract', 'date', 'section', 'region', 'url', 'platform', 'content']
                writer = csv.DictWriter(export_file, fieldnames=header)
                writer.writeheader()
                writer.writerows(parsed_data)
            else:
                json.dump(parsed_data, export_file, ensure_ascii=False, indent=4)



import re
import time
from datetime import datetime
from itertools import count
from typing import Optional, List, Iterable
from urllib.parse import unquote

from loguru import logger
from requests_html import HTML

from .crawler import Crawler

XPATH_DATA = {
    'news_title': '//h1[@class="c-content-head__title"]/text()|//h1[@itemprop="headline"]/text()|'
                  '//meta[@property="og:title"]/@content|//head/title/text()',
    'news_date': '//meta[@property="article:published_time"]/@content|//time[@itemprop="datePublished"]/@datetime',
    'news_abstract': '//h2[@class="c-content-head__subtitle"]/text()|//h2[@itemprop="alternativeHeadline"]/text()|'
                     '//meta[@property="og:description"]/@content',
    'news_body': '//div[@itemprop="articleBody"]//p/text()|//div[@class="c-news__body"]//p/text()',
    'news_section': '//meta[@property="article:section"]/@content',
    'news_region': '//strong[@class="c-signature__location"]//text()|'
                   '//div[@class="c-signature c-signature--left"]//text()',
    'news_tags': '//meta[@name="keywords"]/@content',
    'news_type': '//meta[@property="og:type"]/@content',
    'news_id': "//section[@id='comentarios']",
}


class FolhaNews(Crawler):
    def __init__(self, use_database: bool = True) -> None:
        super().__init__(use_database=use_database)

        self._SEARCH_API = "https://search.folha.uol.com.br/?q={}&site=todos"

    def _make_request(self, target_url: str) -> Optional[HTML]:
        for _ in range(100):
            try:
                response = self.SESSION.get(url=target_url)
                if response.status_code == 200:
                    html_data = HTML(html=response.content, url=response.url)
                    return html_data

            except self._ERRORS:
                logger.warning(
                    f"Folha de São Paulo servers are trying to break the capture. Waiting 5 seconds before retrying."
                )
                self.SESSION.cookies.clear_session_cookies()
                time.sleep(5)

        return None

    @staticmethod
    def _extract_title(article_page: HTML) -> Optional[str]:
        title = article_page.xpath(XPATH_DATA['news_title'], first=True)
        if title is not None:
            return title

        return None

    @staticmethod
    def _extract_abstract(article_page: HTML) -> Optional[str]:
        abstract = article_page.xpath(XPATH_DATA['news_abstract'], first=True)
        if abstract is not None:
            return abstract

        return None

    @staticmethod
    def _extract_date(article_page: HTML) -> Optional[datetime]:
        raw_date = article_page.xpath(XPATH_DATA['news_date'], first=True)
        if raw_date is not None:
            published_date = datetime.strptime(raw_date, "%Y-%m-%d %H:%M:%S")
            return published_date

        return None

    @staticmethod
    def _extract_section(article_page: HTML) -> Optional[str]:
        section = article_page.xpath(XPATH_DATA['news_section'], first=True)
        if section is not None:
            return section

        return None

    def _extract_region(self, article_page: HTML) -> Optional[str]:
        region = article_page.xpath(XPATH_DATA['news_region'], first=True)
        if region is not None:
            return region

        return None

    @staticmethod
    def _extract_tags(article_page: HTML) -> Optional[str]:
        tags = article_page.xpath(XPATH_DATA['news_tags'], first=True)
        if tags is not None:
            return tags

        return None

    @staticmethod
    def _extract_type(article_data: HTML) -> Optional[str]:
        news_type = article_data.xpath(XPATH_DATA['news_type'], first=True)
        if news_type is not None:
            return news_type

        return None

    @staticmethod
    def _extract_body(article_page: HTML) -> Optional[str]:
        article_body = article_page.xpath(XPATH_DATA['news_body'])
        if article_body is not None:
            body = ' '.join(article_body)
            re.sub(r"\s{2,}", body, "")
            return body

        return None

    @staticmethod
    def _extract_id_data(article_page: HTML) -> Optional[dict]:
        try:
            news_id = dict(article_page.xpath(XPATH_DATA['news_id'], first=True).attrs)
        except AttributeError:
            return None

        if news_id is not None:
            news_id_data = {}
            if "data-section" in news_id.keys():
                news_id_data['category_name'] = news_id['data-section']
            if "data-id" in news_id.keys():
                news_id_data['article_id'] = news_id['data-id']
            if "data-service" in news_id.keys():
                news_id_data['service_name'] = news_id['data-service']
            if "data-type" in news_id.keys():
                news_id_data['data_type'] = news_id['data-type']

            if ('category_name' and 'article_id' and 'service_name' and 'data_type') in news_id_data.keys():
                return news_id_data

        return None

    def parse_news(self, news_urls: list, parse_body: bool = False, save_html: bool = True) -> Iterable[dict]:
        parsed_counter = 0
        for i, url in enumerate(news_urls):
            if '1.folha.uol.com.br' not in url:
                continue

            page = self._make_request(target_url=url)
            logger.info(f"Article {i+1} >> Parsing data at {datetime.now()}.")

            parsed_news = {
                'title': self._extract_title(article_page=page),
                'abstract': self._extract_abstract(article_page=page),
                'date': self._extract_date(article_page=page),
                'section': self._extract_section(article_page=page),
                'region':  self._extract_region(article_page=page),
                'url': url,
                'platform': 'Folha de São Paulo',
                'tags': self._extract_tags(article_page=page),
                'type': self._extract_type(article_data=page),
                'body': self._extract_body(article_page=page),
                'id_data': self._extract_id_data(article_page=page),
                'html': page.raw_html if save_html else None,
            }

            if self.DB.check_duplicates(parsed_data=parsed_news):
                continue

            parsed_counter += 1
            logger.success(f"Article {i + 1} >> Data parsed successfully!.")

            yield parsed_news

        logger.success(
            f"All the data have been parsed successfully! "
            f"{parsed_counter} of {len(news_urls)} news had the data extracted."
        )

    def search_news(self, keywords: list, max_pages: int = -1) -> List[str]:
        news_urls = []
        for keyword in keywords:
            logger.info(f"Retrieving news from Folha de São Paulo associated with the Keyword \"{keyword}\".")
            page = self._make_request(self._SEARCH_API.format(keyword))
            if page is None:
                continue

            for i in count():
                if not page.xpath('//div[@class="c-headline__content"]'):
                    break

                logger.info(
                    f"{keyword.title()} >> Getting data from Page {i+1}. Current URLs acquired: {len(news_urls)}."
                )
                news_list = [url for url in page.xpath(
                    '//div[@class="c-headline__content"]/a/@href'
                )]
                news_urls += (url for url in news_list if '1.folha.uol.com.br' in url)

                if i+1 == max_pages:
                    break

                if not page.xpath('//li[@class="c-pagination__arrow"]'):
                    break

                next_page = unquote(page.xpath('//li[@class="c-pagination__arrow"]/a/@href', first=True))
                page = self._make_request(target_url=next_page)
                if page is None:
                    break

        logger.success(
            f"News retrieved successfully! A total of {len(news_urls)} articles have been found."
        )

        return news_urls



import json.decoder
import time
from datetime import datetime
from itertools import count
from typing import List, Iterable, Optional

import requests.exceptions
import urllib3.exceptions
from loguru import logger
from requests_html import HTMLSession, Element, HTML

from .crawler import Crawler

SESSION = HTMLSession()


class FolhaComments(Crawler):
    def __init__(self) -> None:
        self._XPATH = {
            'comment_items': '//li[@class="c-list-comments__item"]',
            'comment_author': '//strong[@class="c-list-comments__user"]/text()',
            'comment_date': '//time[@class="c-list-comments__date"]/@datetime',
            'comment_text': '//p[@class="c-list-comments__comment"]/text()',
            'comment_upvote': '//button[@class="c-list-comments__rating"]/span/text()',
            'comment_id': '//button[@class="c-list-comments__rating"]/@data-comment-rating',
        }
        self._COMMENTS_API = 'https://comentarios1.folha.uol.com.br/comentarios/{}?sr={}'
        self._COMMENTS_ENGINE = 'https://comentarios1.folha.uol.com.br/comentarios.jsonp'
        self._ERRORS = (
            requests.exceptions.ReadTimeout, requests.exceptions.InvalidSchema, requests.exceptions.MissingSchema,
            urllib3.exceptions.ConnectionError, urllib3.exceptions.ProtocolError, ConnectionResetError,
            requests.exceptions.ConnectionError, requests.exceptions.ChunkedEncodingError
        )

    def _make_request(self, target_url: str, payload: dict = None, request_data: dict = None) -> Optional[HTML]:
        for _ in range(100):
            try:
                if (payload and request_data) is None:
                    response = SESSION.get(url=target_url)
                else:
                    response = SESSION.get(url=target_url, params=payload, cookies=request_data)

                if response.status_code == 200:
                    html_data = HTML(html=response.content, url=response.url)
                    return html_data

            except self._ERRORS:
                logger.warning(
                    f"Folha de São Paulo servers are trying to break the capture. Waiting 5 seconds before retrying."
                )
                SESSION.cookies.clear_session_cookies()
                time.sleep(5)

        return None

    @staticmethod
    def _gen_pagination() -> Iterable[int]:
        for page_index in count():
            logger.info(f"Getting comments from Page {page_index+1}.")
            page_number = 1 + (page_index * 50)
            yield page_number

    def _get_api_id(self, news_id_data: dict) -> Optional[int]:
        payload = {
            'service_name': news_id_data['service_name'],
            'type': news_id_data['data_type'],
            'limit': '1',
            'show_replies': 'false',
            'show_with_alternate': 'false',
            'link_format': 'html',
            'order_by': 'create',
            'callback': 'get_comments',
            'category_name': news_id_data['category_name'],
            'external_id': news_id_data['article_id'],
        }

        access_data = {
            'folha_ga_userType': 'not_logged',
            'folha_ga_loginType': 'not_logged',
            'folha_ga_userGroup': 'visitor',
            'folha_ga_swgt': 'sub_na',
        }

        response = self._make_request(target_url=self._COMMENTS_ENGINE, payload=payload, request_data=access_data)
        try:
            raw_data = response.html.replace("get_comments( ", "").replace(" ) ;", "")
            data = json.loads(raw_data)

            api_id = int(data['subject']['subject_id'])

            return api_id
        except (ValueError, KeyError, json.decoder.JSONDecodeError):
            return None

    def _get_comment_data(self, news_id: int) -> Iterable[Element]:
        logger.warning(f"Starting data extraction for comments from Article ID {news_id}.")
        total_comments = 0
        for page in self._gen_pagination():
            target_url = str(self._COMMENTS_API).format(news_id, page)

            response = self._make_request(target_url=target_url)
            if not response.xpath(self._XPATH["comment_items"]):
                break

            comments = response.xpath(self._XPATH["comment_items"])
            total_comments += len(comments)
            logger.info(f"Current number of comments acquired from ID {news_id}: {total_comments}.")
            yield from comments

        logger.success(
            f"A total of {total_comments} comments have been extracted from ID {news_id}! Finished at {datetime.now()}."
        )

    def _extract_author(self, comment_node: Element) -> Optional[str]:
        author = comment_node.xpath(self._XPATH['comment_author'], first=True)
        if author is not None:
            return author

        return None

    def _extract_date(self, comment_node: Element) -> Optional[datetime]:
        raw_date = comment_node.xpath(self._XPATH['comment_date'], first=True)
        if raw_date is not None:
            published_date = datetime.strptime(raw_date, "%Y-%m-%d %H:%M:%S")
            return published_date

        return None

    def _extract_upvote(self, comment_node: Element) -> Optional[int]:
        target_data = comment_node.xpath(self._XPATH['comment_upvote'], first=True)
        if target_data is not None:
            upvote = int(target_data)
            return upvote

        return None

    def _extract_comment_text(self, comment_node: Element) -> Optional[str]:
        comment_text = comment_node.xpath(self._XPATH['comment_text'], first=True)
        if comment_text is not None:
            return comment_text

        return None

    def _extract_comment_id(self, comment_node: Element) -> Optional[int]:
        comment_id = comment_node.xpath(self._XPATH['comment_id'], first=True)
        if comment_id is not None:
            return int(comment_id)

        return None

    def parse_comments(self, news_list: List[dict]) -> Iterable[dict]:
        for news_data in news_list:
            logger.info(f"Checking if \"{news_data['title']}\" have comments.")
            id_data = news_data["id_data"]
            if id_data is None:
                logger.warning(f"No comments found. Proceeding to the next one.")
                continue

            news_id = self._get_api_id(news_id_data=id_data)
            if news_id is None:
                logger.warning(f"No comments found. Proceeding to the next one.")
                continue

            raw_data = [data for data in self._get_comment_data(news_id=news_id)]
            if len(raw_data) == 0:
                continue

            for comment in raw_data:
                if comment is None:
                    continue

                data = {
                    "author": self._extract_author(comment_node=comment),
                    "date": self._extract_date(comment_node=comment),
                    "upvote": self._extract_upvote(comment_node=comment),
                    "news_data": {
                        "title": news_data["title"],
                        "region": news_data["region"],
                        "news_id": news_data["id_data"]["article_id"],
                        "api_id": news_id,
                        "api_url": f"https://comentarios1.folha.uol.com.br/comentarios/{news_id}",
                        "url": news_data["url"],
                    },
                    "comment": self._extract_comment_text(comment_node=comment),
                    "comment_id": self._extract_comment_id(comment_node=comment),
                    "platform": news_data["platform"],
                }

                yield data



import time
from datetime import datetime
from itertools import count
from typing import Optional, List, Iterable

from loguru import logger
from requests_html import HTML

from .crawler import Crawler

XPATH_DATA = {
    'news_abstract': '//meta[@property="og:description"]/@content|//meta[@name="description"]/@content',
    'news_body': '//div[@id="news-body"]',
    'news_type': '//meta[@property="og:type"]/@content',
}


class ExameNews(Crawler):
    def __init__(self, use_database: bool = True) -> None:
        super().__init__(use_database=use_database)

        self._SEARCH_API = "https://content-api.exame.com/api/xm/wp/v2/news"

    def _get_article(self, article_url: str) -> Optional[HTML]:
        for _ in range(100):
            try:
                response = self.SESSION.get(url=article_url)
                if response.status_code == 200:
                    html_data = HTML(html=response.content, url=article_url)
                    return html_data

            except self._ERRORS:
                logger.warning(
                    f"Error while getting article with URL: {article_url}."
                )

        return None

    def _make_search(self, page: int, keyword: str) -> Optional[List[dict]]:
        search_params = {
            'page': f"{page}",
            'per_page': '25',
            '_details': 'true',
            '_fields': ('id,slug,date,link,title,featured_media_url,'
                        'categories_data,sponsor_type,sponsor_name,sponsor_link,acf'),
            'search': keyword,
            'order': 'desc',
        }

        for _ in range(100):
            try:
                response = self.SESSION.get(url=self._SEARCH_API, params=search_params)
                if response.status_code != 200:
                    return None

                search_data = response.json()
                return search_data
            except self._ERRORS:
                logger.warning(
                    f"Exame servers are trying to break the capture. Waiting 5 seconds before retrying."
                )
                self.SESSION.cookies.clear_session_cookies()
                time.sleep(5)

        return None

    @staticmethod
    def _extract_title(article_data: dict) -> Optional[str]:
        title = article_data['title']
        if title is not None:
            return title

        return None

    @staticmethod
    def _extract_abstract(article_page: HTML) -> Optional[str]:
        abstract = article_page.xpath(XPATH_DATA['news_abstract'], first=True)
        if abstract is not None:
            return abstract

        return None

    @staticmethod
    def _extract_date(article_data: dict) -> Optional[datetime]:
        raw_date = article_data['date']
        if raw_date is not None:
            published_date = datetime.strptime(raw_date, "%Y-%m-%dT%H:%M:%S")
            return published_date

        return None

    @staticmethod
    def _extract_section(article_data: dict) -> Optional[str]:
        section = article_data['categories_data'][0]['name']
        if section is not None:
            return section

        return None

    @staticmethod
    def _extract_tags(article_data: dict) -> Optional[str]:
        tags = [tag['name'] for tag in article_data['categories_data']]
        if tags is not None:
            article_tags = ' | '.join(tags)
            return article_tags

        return None

    @staticmethod
    def _extract_type(article_page: HTML) -> Optional[str]:
        news_type = article_page.xpath(XPATH_DATA['news_type'], first=True)
        if news_type is not None:
            return news_type

        return None

    def _extract_region(self, article_page: HTML) -> Optional[str]:
        region = None
        return region

    @staticmethod
    def _extract_body(article_page: HTML) -> Optional[str]:
        article_body = article_page.xpath(XPATH_DATA['news_body'], first=True)
        if article_body is not None:
            body = article_body.full_text
            return body

        return None

    @staticmethod
    def _extract_id(article_data: dict) -> Optional[int]:
        news_id = article_data['id']
        if news_id is not None:
            return news_id

        return None

    def parse_news(self, news_urls: List[dict], parse_body: bool = False, save_html: bool = True) -> Iterable[dict]:
        parsed_counter = 0
        for i, article_data in enumerate(news_urls):
            if article_data is None:
                continue

            url = article_data['link']
            if 'exame.com' not in url:
                continue

            page = self._get_article(article_url=url)
            logger.info(f"Article {i+1} >> Parsing data at {datetime.now()}.")

            parsed_news = {
                'title': self._extract_title(article_data=article_data),
                'abstract': self._extract_abstract(article_page=page),
                'date': self._extract_date(article_data=article_data),
                'section': self._extract_section(article_data=article_data),
                'region':  self._extract_region(article_page=page),
                'url': url,
                'platform': 'Exame',
                'tags': self._extract_tags(article_data=article_data),
                'type': self._extract_type(article_page=page),
                'body': self._extract_body(article_page=page),
                'id_data': self._extract_id(article_data=article_data),
                'html': page.raw_html if save_html else None,
            }

            if self.DB.check_duplicates(parsed_data=parsed_news):
                continue

            parsed_counter += 1
            logger.success(f"Article {i + 1} >> Data parsed successfully!.")

            yield parsed_news

        logger.success(
            f"All the data have been parsed successfully! "
            f"{parsed_counter} of {len(news_urls)} news had the data extracted."
        )

    def search_news(self, keywords: list, max_pages: int = -1) -> List[dict]:
        articles = []
        for keyword in keywords:
            logger.info(f"Retrieving news from Exame associated with the Keyword \"{keyword}\".")

            for i in count():
                logger.info(
                    f"{keyword.title()} >> Getting data from Page {i+1}. Current URLs acquired: {len(articles)}."
                )

                search_data = self._make_search(page=i+1, keyword=keyword)
                if search_data is None:
                    logger.success(
                        f"{keyword.title()} >> All data have been added to list! Finished at {datetime.now()}."
                    )
                    break

                for item in search_data:
                    if "link" not in item.keys():
                        continue

                    if "exame.com" in item["link"]:
                        articles.append(item)

                if i+1 == max_pages:
                    logger.success(
                        f"{keyword.title()} >> All data have been added to list! Finished at {datetime.now()}."
                    )
                    break

        logger.success(
            f"News retrieved successfully! A total of {len(articles)} articles have been found."
        )

        return articles



import csv
import json
import traceback
from datetime import datetime
from typing import List, Iterable, Optional

import pymongo
import pymongo.database
import pymongo.errors
from loguru import logger


class PyBrNewsDB:
    """
    pyBrNews Database Class to use MongoDB database to store all article and comment data from the platforms.

    By default, uses the standard MongoDB host and port (localhost:27017). Can be changed with set_connection,
    with the parameters host and port.

    Example: set_connection(host="192.168.0.1", port: 88890).
    """
    def __init__(self, data_kind: str = "news") -> None:
        if "news" not in data_kind and "comments" not in data_kind:
            raise ValueError(
                f"An invalid kind of data for database [ {data_kind} ] was supplied. Review and try again."
            )

        self.client: Optional[pymongo.MongoClient] = None
        self.db: Optional[pymongo.database.Database] = None

        self.set_connection()
        self.collection = self.db.get_collection(data_kind)

    def set_connection(self, host: str = "localhost", port: int = 27017) -> None:
        """
        Sets the connection host:port parameters for the MongoDB. By default, uses the standard localhost:27017 for
        local usage.

        Parameters:
             host (str): Hostname or address to connect.
             port (int): Port to be used in the connection.
        """
        self.client = pymongo.MongoClient(host=host, port=port)
        self.db = self.client.get_database(name="pyBrNews")

    def insert_data(self, parsed_data: dict) -> None:
        """
        Inserts the parsed data from a news article or extracted comment into the DB Backend (MongoDB - pyMongo).

        Parameters:
            parsed_data (dict): Dictionary containing the parsed data from a news article or comment.
        Returns:
            None: Shows a success message if the insertion occurred normally. If not, shows an error message.
        """
        parsed_data["entry_dt"] = datetime.now()

        try:
            inserted_data = self.collection.insert_one(parsed_data)
            logger.success(
                f"Data inserted into pyBrNews DB! Document ID {inserted_data.inserted_id} "
                f"successfully added to the news collection."
            )
        except Exception as e:
            logger.error(f"An error happened while attempting to insert the given data to the pyBrNews database.")
            logger.debug(f"Data URL: {parsed_data['url']}")
            logger.debug(f"{traceback.print_exception(e)}")

    def check_duplicates(self, parsed_data: dict) -> bool:
        """
        Checks if the parsed data is already in the database and prevents from being duplicated
        in the crawler execution.

        Parameters:
            parsed_data (dict): Dictionary containing the parsed data from a news article or comment.
        Returns:
            bool: True if the given parsed data is already in the database. False if not.
        """
        check_params = {
            "url": parsed_data["url"],
            "date": parsed_data["date"],
        }

        documents = self.collection.find_one(check_params)
        if documents is None:
            return False

        return True


class PyBrNewsFS:
    """
    pyBrNews File System Class for exporting data without using the MongoDB database. Allows to export parsed data into
    individual JSON or condenses all the parsed data into a CSV file for external usage.

    By default, exports all the files into the current work directory. To alter the save path, call the set_save_path()
    method, passing the attribute "fs_save_path" with the desired directory ending with a slash.
    """
    def __init__(self) -> None:
        self.save_path = ""

    def set_save_path(self, fs_save_path: str) -> None:
        """
        Sets the save path for all the exported data generated by this Class.

        Example: set_save_path(fs_save_path="/home/ubuntu/newsData/")

        Parameters:
             fs_save_path (str): Desired save path directory, ending with a slash.
        """
        if "/" not in fs_save_path[-1]:
            raise ValueError("End the save path with a slash ( / ).")

        self.save_path = fs_save_path

    def to_json(self, parsed_data: dict) -> None:
        """
        Using the parsed data dictionary from a news article or a comment, export the data as an individual JSON file.

        Parameters:
            parsed_data (dict): Dictionary containing the parsed data from a news article or a comment.
        """
        if parsed_data is None:
            raise AttributeError("Parsed Data Dictionary cannot be an NoneType value.")

        export_time = datetime.today().strftime("%Y_%m_%d_%H_%M_%S")
        file_name = f"ParsedNewsData_{export_time}.json"
        try:
            with open(f"{self.save_path}{file_name}", mode="w", encoding="utf-8") as json_file:
                json.dump(parsed_data, json_file, ensure_ascii=False, indent=4)

            logger.success(
                f"Data saved successfully as a JSON file! Document path: {self.save_path}{file_name}"
            )
        except OSError:
            logger.error(
                "An error occurred while attempting to save the JSON file. Review the save path and the data, and try "
                "again."
            )

    @staticmethod
    def _remove_null_values(raw_full_data: List[dict]) -> Iterable[dict]:
        """
        By a given list of dictionaries containing the parsed data from news or comments, yields per iteration the dict
        only if the item is not None.

        Parameters:
            raw_full_data (List[dict]): List containing the original dictionaries of parsed data (incl. None).
        Returns:
            Iterable[dict]: Per iteration -> If the item is not None, yields the parsed data dictionary.
        """
        for data in raw_full_data:
            if data is not None:
                return data

    @staticmethod
    def check_duplicates(parsed_data: dict = None) -> bool:
        """
        Placeholder method, mirroring the existing one in the PyBrNewsDB Class. Just returns False because is a FS data
        structure.

        Parameters:
            parsed_data (dict): Dictionary containing the parsed data from a news article or comment.
        Returns:
            bool: Because it is a FS Class, it does not check for duplicates. Returns False.
        """
        if parsed_data is not None:
            logger.warning("PyBrNews File System in use. Checking for duplicates only works on the PyBrNews Database.")

        return False

    def export_all_data(self, full_data: List[dict]) -> None:
        """
        By a given list of dictionaries containing the parsed data from news or comments, export in a CSV file
        containing all data.

        Parameters:
            full_data (List[dict]): List containing the dictionaries of parsed data.
        """
        export_time = datetime.today().strftime("%Y_%m_%d_%H_%M_%S")
        export_data = [data for data in self._remove_null_values(raw_full_data=full_data)]

        if len(full_data) == 0:
            raise ValueError(f"An empty list of parsed data cannot be used.")

        header = [
            "title", "abstract", "date", "section", "region", "url", "platform",
            "tags", "type", "body", "id_data", "html",
        ]

        file_name = f"ParsedNewsData_{export_time}.csv"
        try:
            with open(f"{self.save_path}{file_name}", mode="w", encoding="utf-8") as export_file:
                writer = csv.DictWriter(export_file, fieldnames=header)
                writer.writeheader()
                writer.writerows(export_data)

            logger.success(
                f"All data exported successfully as a CSV file! Document path: {self.save_path}{file_name}"
            )
        except OSError:
            logger.error(
                "An error occurred while attempting to save the CSV file. Review the save path and the data, and try "
                "again."
            )



import codecs

from setuptools import find_packages, setup

with codecs.open("README_ENG.md", "r", encoding='utf-8') as readme_file:
    readme = readme_file.read()

setup(
    name='pyBrNews',
    packages=find_packages(),
    install_requires=[
        "loguru>=0.6.0", "pymongo>=4.3.2", "requests_html>=0.10.0"
    ],
    version='0.1.2',
    description='A Brazilian News Website Data Acquisition Library for Python',
    long_description=readme,
    long_description_content_type="text/markdown",
    url="https://www.github.com/NepZR/pyBrNews/",
    author='Lucas Darlindo Freitas Rodrigues',
    author_email="lucas.darlindo@gmail.com",
    license='GNU GPLv3',
    license_files=(
        'LICENSE',
    ),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Framework :: IPython",
        "Framework :: Jupyter",
        "Framework :: Jupyter :: JupyterLab",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Information Technology",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Natural Language :: English",
        "Natural Language :: Portuguese (Brazilian)",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Database",
        "Topic :: Education",
        "Topic :: Internet",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Version Control :: Git",
        "Topic :: Text Processing",
        "Topic :: Text Processing :: Markup :: HTML",
        "Topic :: Text Processing :: Markup :: XML"
    ]
)



import json
import re
from datetime import datetime
from itertools import count
from typing import List, Iterable, Optional
from urllib.parse import unquote

from loguru import logger
from requests_html import HTML

from .crawler import Crawler
from ..config import g1_api

XPATH_DATA = {
    'news_title': '//div[@class="title"]/h1/text()|//meta[@name="title"]/@content|//head/title/text()',
    'news_date': '//time[@itemprop="datePublished"]/@datetime',
    'news_abstract': '//h2[@class="content-head__subtitle"]/text()|//meta[@name="description"]/@content',
    'news_body': '//div[@class="mc-article-body"]//p/text()',
    'news_section': '//div[@class="header-title-content"]/a/text()',
    'news_tags': '//ul[@class="entities__list"]//a/text()'
}


class G1News(Crawler):
    def __init__(self, use_database: bool = True) -> None:
        super().__init__(use_database=use_database)

        self._API_CONFIG = g1_api.news_config
        self._NEWS_API = self._API_CONFIG['api_url']['news_engine']
        self._SEARCH_API = self._API_CONFIG['api_url']['search_engine']

    def _retrieve_news_by_region(self, regions: list, max_pages: int = -1) -> Iterable[str]:
        for region in regions:
            for i in count():
                try:
                    page = self.SESSION.get(
                        self._NEWS_API.format(self._API_CONFIG['regions'][region], str(i + 1))
                    ).json()
                except json.decoder.JSONDecodeError:
                    break

                for item in page['items']:
                    try:
                        url = str(item['content']['url']) if ('materia' in item['type']) else None
                        if url is not None:
                            logger.success(
                                f"URL from G1 {region.upper()} retrieved successfully! Item added to list: {url}"
                            )
                            yield url
                    except KeyError:
                        for article in item['content']['posts']:
                            url = article['url']
                            if url is not None:
                                logger.success(
                                    f"URL from G1 {region.upper()} retrieved successfully! Item added to list: {url}"
                                )
                                yield url

                if i+1 == max_pages:
                    break

    def _retrieve_news_brazil(self, max_pages: int = -1) -> Iterable[str]:
        for i in count():
            try:
                page = self.SESSION.get(
                    self._NEWS_API.format(self._API_CONFIG['regions']['brasil'], str(i + 1))
                ).json()
            except json.decoder.JSONDecodeError:
                break

            for item in page['items']:
                try:
                    url = item['content']['url'] if ('materia' in item['type']) else ()
                    if url is not None:
                        logger.success(f"URL from G1 retrieved successfully! Item added to list: {url}")
                        yield url
                except KeyError:
                    for article in item['content']['posts']:
                        url = article['url']
                        if url is not None:
                            logger.success(
                                f"URL from G1 retrieved successfully! Item added to list: {url}"
                            )
                            yield url

            if i+1 == max_pages:
                break

    def retrieve_latest_news(self, regions: list = None, max_pages: int = -1) -> List[str]:
        news_urls = []
        if regions is None:
            news_urls = [url for url in self._retrieve_news_brazil(max_pages=max_pages)]

        if regions is not None:
            news_urls = [url for url in self._retrieve_news_by_region(regions=regions, max_pages=max_pages)]

        return news_urls

    def parse_news(self, news_urls: List[str], parse_body: bool = False, save_html: bool = True) -> Iterable[dict]:
        parsed_counter = 0
        for i, url in enumerate(news_urls):
            logger.info(f"Article {i+1} >> Parsing data at {datetime.now()}.")
            page = self.SESSION.get(url).html

            parsed_news = {
                'title': self._extract_title(article_page=page),
                'abstract': self._extract_abstract(article_page=page),
                'date': self._extract_date(article_page=page),
                'section': self._extract_section(article_page=page),
                'region': self._extract_region(article_url=url),
                'url': url,
                'platform': 'Portal G1',
                'tags': self._extract_tags(article_data=page),
                'type': self._extract_type(article_url=url),
                'body': self._extract_body(article_page=page),
                'id_data': None,
                'html': page.raw_html if save_html else None,
            }

            if self.DB.check_duplicates(parsed_data=parsed_news):
                continue

            parsed_counter += 1
            logger.success(f"Article {i + 1} >> Data parsed successfully!.")

            yield parsed_news

        logger.success(
            f"All the data have been parsed successfully! "
            f"{parsed_counter} of {len(news_urls)} news had the data extracted."
        )

    def search_news(self, keywords: List[str], max_pages: int = -1) -> List[str]:
        news_urls = []
        for keyword in keywords:
            logger.info(f"Retrieving news from G1 associated with the Keyword \"{keyword}\".")
            for i in count():
                page = self.SESSION.get(self._SEARCH_API.format(keyword, str(i+1)))
                if "page" not in page.url:
                    break

                news_list = [unquote(url).split('u=')[1].split('&')[0] for url in page.html.xpath(
                    '//div[@class="widget--info__text-container"]/a/@href'
                )]
                news_urls += (url for url in news_list if 'g1.globo.com' in url)

                if i+1 == max_pages:
                    break

        logger.success(
            f"News retrieved successfully! A total of {len(news_urls)} articles have been found."
        )

        return news_urls

    @staticmethod
    def _extract_title(article_page: HTML) -> Optional[str]:
        title = article_page.xpath(XPATH_DATA['news_title'], first=True)
        if title is not None:
            return title

        return None

    @staticmethod
    def _extract_abstract(article_page: HTML) -> Optional[str]:
        abstract = article_page.xpath(XPATH_DATA['news_abstract'], first=True)
        if abstract is not None:
            return abstract

        return None

    @staticmethod
    def _extract_date(article_page: HTML) -> Optional[datetime]:
        raw_date = article_page.xpath(XPATH_DATA['news_date'], first=True)
        if raw_date is not None:
            try:
                published_date = datetime.strptime(raw_date, "%Y-%m-%dT%H:%M:%S.%fZ")
                return published_date
            except ValueError:
                return None

        return None

    @staticmethod
    def _extract_section(article_page: HTML) -> Optional[str]:
        section = article_page.xpath(XPATH_DATA['news_section'], first=True)
        if section is not None:
            return section

        return None

    def _extract_region(self, article_url: str) -> Optional[str]:
        region = article_url.split('/')[3]
        if region in self._API_CONFIG['regions'].keys():
            region = region.upper()
            return region

        return None

    @staticmethod
    def _extract_tags(article_data: HTML) -> Optional[str]:
        tags = article_data.xpath(XPATH_DATA['news_tags'], first=True)
        if tags is not None:
            return tags

        return None

    @staticmethod
    def _extract_type(article_url: str) -> Optional[str]:
        if "video" in article_url:
            news_type = "Video"
        else:
            news_type = "Article"

        return news_type

    @staticmethod
    def _extract_body(article_page: HTML) -> Optional[str]:
        article_body = article_page.xpath(XPATH_DATA['news_body'])
        if article_body is not None:
            body = ' '.join(article_body)
            re.sub(r"(\s{2,})|(\n)+", body, "")
            return body

        return None



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Optional, List, Union, Iterable

import requests.exceptions
import urllib3.exceptions
from requests_html import HTMLSession, HTML

from ..config.database import PyBrNewsDB, PyBrNewsFS


class Crawler(ABC):
    def __init__(self, use_database: bool = True) -> None:
        self.SESSION = HTMLSession()

        self.DB: Union[PyBrNewsDB, PyBrNewsFS]
        if not use_database:
            self.DB = PyBrNewsFS()
        else:
            self.DB = PyBrNewsDB()

        self._ERRORS = (
            requests.exceptions.ReadTimeout, requests.exceptions.InvalidSchema, requests.exceptions.MissingSchema,
            urllib3.exceptions.ConnectionError, urllib3.exceptions.ProtocolError, ConnectionResetError,
            requests.exceptions.ConnectionError, requests.exceptions.ChunkedEncodingError
        )

    @abstractmethod
    def parse_news(self,
                   news_urls: List[Union[str, dict]],
                   parse_body: bool = False,
                   save_html: bool = True) -> Iterable[dict]:
        """
        Extracts all the data from the article in a given news platform by iterating over a URL list. Yields a
        dictionary containing all the parsed data from the article.

        Parameters:
            news_urls (List[str]): A list containing all the URLs or a data dict to be parsed from a given platform.
            parse_body (bool): Defines if the article body will be extracted.
            save_html (bool): Defines if the HTML bytes from the article will be extracted.
        Returns:
             Iterable[dict]: Dictionary containing all the article parsed data.
        """
        pass

    @abstractmethod
    def search_news(self,
                    keywords: List[str],
                    max_pages: int = -1) -> List[Union[str, dict]]:
        """
        Extracts all the data or URLs from the news platform based on the keywords given. Returns a list containing the
        URLs / data found for the keywords.

        Parameters:
            keywords (List[str]): A list containing all the keywords to be searched in the news platform.
            max_pages (int): Number of pages to have the articles URLs extracted from.
                             If not set, will catch until the last possible.
        Returns:
             List[Union[str, dict]]: List containing all the URLs / data found for the keywords.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_title(article_page: HTML) -> Optional[str]:
        """
        Extracts the title of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The title of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_abstract(article_page: HTML) -> Optional[str]:
        """
        Extracts the abstract of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The abstract of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_date(article_page: HTML) -> Optional[datetime]:
        """
        Extracts the date of a given news article page and returns as a datetime for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[datetime]: The published date of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_section(article_page: HTML) -> Optional[str]:
        """
        Extracts the section of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The section name of the news article. None if not found.
        """
        pass

    @abstractmethod
    def _extract_region(self, article_page: HTML) -> Optional[str]:
        """
        Extracts the region of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The region of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_tags(article_data: Union[HTML, str]) -> Optional[str]:
        """
        Extracts the tags/keywords of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_data (Union[HTML, str]): A HTML or string object containing the data from the news article page.
        Returns:
            Optional[str]: The tags/keywords of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_type(article_page: HTML) -> Optional[str]:
        """
        Extracts the type of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The type of the news article. None if not found.
        """
        pass

    @staticmethod
    @abstractmethod
    def _extract_body(article_page: HTML) -> Optional[str]:
        """
        Extracts the full body text of a given news article page and returns as a string for the parsed data dictionary.

        Parameters:
             article_page (HTML): A HTML object containing the data from the news article page.
        Returns:
            Optional[str]: The full body text of the news article. None if not found.
        """
        pass
